from PIL import Image
from datetime import datetime

data = {}

def mandelbrot(args):
    row = args[0]
    col = args[1]
    size_x = args[2]
    size_y = args[3]
    # data = args[4]
    global data

    ret = [ 0.0 ] *3

    x0 = float(row) / size_x * 3.5 - 2.5
    y0 = float(col) / size_y * 2.0 - 1.0
    assert -2.5 <= x0 <= 1
    assert -1 <= y0 <= 1
    x = 0
    y = 0
    iteration = 0
    max_iteration = 1000
    while (x*x + y*y) < 4 and (iteration < max_iteration):
        xtemp = x*x - y*y + x0
        y = 2*x*y + y0
        x = xtemp
        iteration = iteration + 1
        color = iteration % 255
        #print(iteration, color)
        data[(row, col, 0)]  = color 
        data[(row, col, 1)] = (color + 75) % 255
        data[(row, col, 2)] = (color + 150) % 255
        # arr[(row, col, 0)] = color
        # arr[(row, col, 1)] = (color + 75) % 255
        # arr[(row, col, 2)] = (color + 150) % 255
    # return (ret[0], ret[1], ret[2])

def save_png(filename, data_map, size_x, size_y):
    im = Image.new('RGB', (size_x, size_y))
    pixels1 = im.load()
    for i in xrange(size_x):
        for j in xrange(size_y):
                pixels1[i, j] = (data_map[(i, j, 0)], data_map[(i,j,1)], data_map[(i,j,2)])
    im.save(filename, "PNG")

def par_for(f, *args):
    from itertools import product
    from multiprocessing import Pool
    p = Pool(4)

    p.map(f,[e for e in product(*args)])

    # def ret(l):
    #     from multiprocessing import Pool
    #     p = Pool(4)
    #     tmp = p.map(f, l)
    #     return tmp
    # return ret

def fast_mandelbrot(size_x, size_y):
    # from multiprocessing import Pool
    # p = Pool(4)
    # tmp = p.map(mandelbrot, [ (i, j, size_x, size_y) for i in xrange(size_x) for j in xrange(size_y) ])
    # data_map = {}
    # for i in xrange(size_x):
    #     for j in xrange(size_y):
    #         data_map[(i, j)] = tmp[i*size_y +j]
    # return data_map
    # tmp = par(mandelbrot)([(i, j, size_x, size_y) for i in range(size_x) for j in range(size_y)])
    
    # data_map = {}
    # for i in xrange(size_x):
    #     for j in xrange(size_y):
    #         data_map[(i, j)] = tmp[i*size_y +j]
    # return data_map    
    par_for(mandelbrot, xrange(size_x), xrange(size_y), [size_x], [size_y])
    
    # print data_map

    # return data_map

def slow_mandelbrot(size_x, size_y):
    # arr = numpy.zeros(dtype=numpy.int8, shape=(size_x, size_y, 3))
    # arr = [[0, 0, 0] for i in range()]
    data = {}
    for row in range(size_x):
        for col in range(size_y):
            mandelbrot((row, col, size_x, size_y))

        # if row == col:
        # arr[row, col, 0] = 255
        # return arr    
        # for x in []:
            # x0 = float(row) / size_x * 3.5 - 2.5
            # y0 = float(col) / size_y * 2.0 - 1.0
            # assert -2.5 <= x0 <= 1
            # assert -1 <= y0 <= 1
            # x = 0
            # y = 0
            # iteration = 0
            # max_iteration = 1000
            # while (x*x + y*y) < 4 and (iteration < max_iteration):
            #     xtemp = x*x - y*y + x0
            #     y = 2*x*y + y0
            #     x = xtemp
            #     iteration = iteration + 1
            #     color = iteration % 255
            #     #print(iteration, color)
            #     arr[(row, col, 0)] = color
            #     arr[(row, col, 1)] = (color + 75) % 255
            #     arr[(row, col, 2)] = (color + 150) % 255
    return data

if __name__ == "__main__":
    size_x, size_y = 500, 500
    filename = "mandelbrot.png"
    # slow version
    start = datetime.now()
    # data_map = slow_mandelbrot(size_x, size_y)
    slow_mandelbrot(size_x, size_y)
    data_map = data
    end = datetime.now()
    print "slow: runtime: %s" % str(end-start)
    save_png(filename, data_map, size_x, size_y)

    # multi-process version
    start = datetime.now()
    # data_map = fast_mandelbrot(size_x, size_y)
    fast_mandelbrot(size_x, size_y)
    data_map = data
    end = datetime.now()
    print "fast: runtime: %s" % str(end-start)
    save_png("mp-"+filename, data_map, size_x, size_y)    



import re 

class SharedVariables:
    def __init__(self):
        self.shared = {}

    def register(self, variable):
        self.shared[variable] = variable

    def dump(self):
        return self.shared.keys()

    def clean(self):
        self.shared = {}

class Parser:
    space_prog = re.compile("^( *)[^ ]*")
    comment_prog = re.compile("(#.*)$")
    for_prog = re.compile("for ([^ ]+) in ([^:]+):")

    def __init__(self, inputfilename):
        self.f = open(inputfilename)
        self.last_line = None
        self.last_struct = None

    def parse_for(self, line):
        tmp = Parser.for_prog.search(line)

        if tmp is None:
            return None, None
        else:
            line, line_struct = self.parse_line(line)

            return line, {'indention' : line_struct['indention'], 
                            "variable": tmp.group(1), 
                            "iterate_list": tmp.group(2)}

    def parse_blcok(self):
        block_lines = []
        last_indention = None
        while True:
            line, struct = self.parse()
            if line is None:
                break

            if last_indention is None:
                last_indention = struct['indention']
                block_lines.append(line)
            else:
                if last_indention > struct['indention']: # next block
                    self.last_line, self.last_struct = line, struct
                    break
                else:
                    block_lines.append(line)

        return block_lines, {'indention': last_indention}

    def parse(self):
        if self.last_line is not None:
            tmp_line = self.last_line
            tmp_struct = self.last_struct
            self.last_line = None
            self.last_struct = None
            return tmp_line, tmp_struct

        self.line = self.f.readline()

        # last line
        if self.line == '':
            return None, None  

        self.line = self.line[:-1]

        return self.parse_line(self.line)

    def parse_line(self, line):
        # indention 
        space_match = Parser.space_prog.match(self.line)
        indention = 0 
        if space_match is not None:
            indention = len(space_match.group(1))

        # comment
        comment_match = Parser.comment_prog.match(self.line)
        comment = None
        if comment_match is not None:
            comment = comment_match.group(1)

        # tokens :
        tokens = []
        split_tmp = [t for t in self.line.split(" ") if len(t) > 0]
        pragma_tokens = None
        for i in xrange(len(split_tmp)):
            if len(split_tmp[i]) == 0:
                continue
            if split_tmp[i] == Annotation.PREFIX:
                pragma_tokens = split_tmp[i:]
                break
            if split_tmp[i] == "#":
                break
            tokens.append(split_tmp[i])

        return self.line, {
                            "indention": indention, 
                            "comment": comment, 
                            "tokens": tokens, 
                            "pragma_tokens": pragma_tokens
                            }
    def close(self):
        self.f.close()

class Writer:
    def __init__(self, outputfilename):
        self.f = open(outputfilename, "w")

    def write(self, line):
        print >> self.f, line

    def write_block(self, block):
        for line in block:
            self.write(line)

    def close(self):
        self.f.close()

class Template:
    @staticmethod
    def def_function(variable, function_name, indention=0):
        return "%sdef %s(%s):" %(" " * indention, function_name, variable)
    @staticmethod
    def define_variable(left, right, indention=0):
        return "%s%s = %s" %(" " * indention, left, right)
    @staticmethod
    def return_variable(variable, indention=0):
        return "%sreturn %s" % (" "*indention, variable)

    @staticmethod
    def import_moudle(module, package=None,abbr=None, indention=0):
        indention_str = " " * indention
        if package is None:
            if abbr is None:
                return "%simport %s" %(indention_str, module)
            else:
                return "%simport %s as %s" %(indention_str, module, abbr)            
        else:
            if abbr is None:
                return "%sfrom %s import %s" %(indention_str, package, module)
            else:
                return "%sfrom %s import %s as %s" %(indention_str, package, module, abbr) 

    @staticmethod
    def execute_function(function_name, variables, indention=0):
        if type(variables) is str:
            variables_str = variables
        elif type(variables) is list:
            variables_str = ", ".join(variables)
        return "%s%s(%s)" % (" "* indention, function_name, variables_str)

class Annotation:
    PREFIX = "#pragma"
    SHARED_VARIABLES = "#pragma shared"
    PARALLEL_FOR = "#pragma omp parallel for"

class STATE:
    NORMAL = 1
    PARALLEL_FOR = 2

class PragmaTranslator:
    parallel_prog = re.compile(Annotation.PARALLEL_FOR)
    shared_variables_prog = re.compile(Annotation.SHARED_VARIABLES)
    PARALLEL_FOR = 1 
    SHARED_VARIABLES = 2
    ELSE = 3

    @staticmethod
    def parse(pragma_tokens):
        command = " ".join(pragma_tokens)
        tmp = PragmaTranslator.parallel_prog.match(command)
        if tmp is not None:
            return PragmaTranslator.PARALLEL_FOR 
        
        tmp = PragmaTranslator.shared_variables_prog.match(command)
        if tmp is not None:
            return PragmaTranslator.SHARED_VARIABLES
        
        return PragmaTranslator.ELSE

class Translator:
    def __init__(self, inputfilename, outputfilename):
        self.parser = Parser(inputfilename)
        self.writer = Writer(outputfilename)
        self.state = STATE.NORMAL
        self.shared_variables = SharedVariables()
        self.threads = 4

    def process_parallel_for(self, line, struct):
        # parallel for 
        parallel_line, parallel_struct = self.parser.parse_for(line)
        if parallel_line is None:
            # cannot parse parallel for correctly
            self.writer.write(line)
        else:
            # parse block 
            block, block_struct = self.parser.parse_blcok()
            # rewrite code
            self.writer.write(
                Template.def_function(
                    variable=parallel_struct['variable'],
                    function_name="core",
                    indention=parallel_struct['indention']
                )
            )
            # insert registered code 
            self.writer.write(
                Template.define_variable(
                    left = "__shared__",
                    right= "{}",
                    indention=block_struct['indention']
                )
            )
            # import copy 
            self.writer.write(
                Template.import_moudle(
                    module="copy",
                    indention=block_struct['indention']
                )
            )

            for registered_variable in self.shared_variables.dump():
                self.writer.write(
                    Template.define_variable(
                        left = "__shared__['%s']" % registered_variable,
                        right = "copy.deepcopy(%s)" % registered_variable,
                        indention=block_struct['indention'] 
                    )
                )

            # insert block, replace register variable with __shared['%s'] 
            block_text = "\n".join(block)
            for registered_variable in self.shared_variables.dump():
                block_text = block_text.replace(registered_variable, "__shared__['%s']" % registered_variable) 
            self.writer.write(block_text)

            # insert footer 
            self.writer.write(
                Template.return_variable(
                    variable="__shared__",
                    indention=block_struct['indention']
                )
            )
            # insert import 
            self.writer.write(
                Template.import_moudle(
                    package="pathos.multiprocessing",
                    module="ProcessingPool",
                    indention=parallel_struct['indention']
                )
            )
            # insert __shared__ = ProcessingPool(4).map(core, range(size_x))
            self.writer.write(
                    Template.define_variable(
                        left = "__shared__",
                        right = "ProcessingPool(%d).map(core, %s)" %(
                            self.threads, 
                            parallel_struct['iterate_list']),
                        indention=parallel_struct['indention'] 
                    )
            )
            # insert Merge function 
            tmp_list =[ ]
            for registered_variable in self.shared_variables.dump():
                tmp_list.append("'%s': %s" % (registered_variable, registered_variable) )     
            registered_variable_dict_str = "{ %s }" %(", ".join(tmp_list))

            self.writer.write(
                Template.execute_function(
                    function_name="join_shared",
                    variables=["__shared__", registered_variable_dict_str],
                    indention=parallel_struct['indention']
                )
            )

    def process_shared_variables(self, line, struct):
        # need to check token struct first
        assert re.search("[^=]+=[^ ]+", line) ==None, "'%s' is not a delaration of variable."

        # parse variable name 
        variable = struct['tokens'][0]

        # register
        self.shared_variables.register(variable)

    def header(self):
        self.writer.write(
            Template.import_moudle(
                package="lazymp.helpers",
                module="join_dict"
            )
        )
        self.writer.write(
            Template.import_moudle(
                package="lazymp.helpers",
                module="join_shared"
            )
        )

    def run(self):
        self.header()

        while True:
            line, struct = self.parser.parse()
            if line == None:
                break

            if self.state == STATE.NORMAL:
                if struct.get('pragma_tokens') is not None:
                    pragma_type = PragmaTranslator.parse(struct.get('pragma_tokens'))
                    if pragma_type == PragmaTranslator.PARALLEL_FOR:
                        self.process_parallel_for(line, struct)
                    elif pragma_type == PragmaTranslator.SHARED_VARIABLES:
                        self.process_shared_variables(line, struct)
                        self.writer.write(line)
                    else:
                        self.writer.write(line)
                else:
                    self.writer.write(line)

    def close(self):
        self.parser.close()
        self.writer.close()

def translate(inputfilename, outputfilename):
    translator = Translator(inputfilename, outputfilename)
    translator.run()
    translator.close()

if __name__ == "__main__":
    import sys
    inputfilename = sys.argv[1]
    outputfilename = sys.argv[2]
    translate(inputfilename, outputfilename)



def join_dict(from_dict, to_dict):
    for k, v in from_dict.items():
        to_dict[k] = v 

def join_shared(shared_variables_list, variables):
    for shared_variables in shared_variables_list:
        for name, value in shared_variables.items():
            if type(value) == dict:
                join_dict(value, variables[name])


from setuptools import setup, find_packages
setup(
    name = "lazymp",
    version = "0.1.0.0.1",
    packages = find_packages(),
    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires = [
        "pyre>=0.8.2.0-pathos", 
        "processing>=0.52-pathos", 
        "Pillow>=2.8.1", 
        "pathos>=0.2a1.dev0", 
        "pyre>=0.8.2.0-pathos", 
        "processing>=0.52-pathos"    
    ],

    dependency_links = [
        "https://github.com/uqfoundation/pathos/raw/master/external/pyre-0.8.2.0-pathos.zip",
        "https://github.com/uqfoundation/pathos/raw/master/external/processing-0.52-pathos.zip",
        "git+https://github.com/uqfoundation/pathos#egg=pathos-0.2a1.dev0"
    ],

    # metadata for upload to PyPI
    author = "San-Chuan Hung",
    author_email = "sanchuah@andrew.cmu.edu",
    description = "This is an Example Package",
    license = "MIT",
    keywords = "easy lazy openmp parallel",
    url = "http://leohung.net/pymp",   # project home page, if any
    # could also include long_description, download_url, classifiers, etc.
)
